# Diversity.py - Provides methods for analyzing species diversity.
#
# Copyright (C) 2009 Jason J. Roberts
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (available in the file LICENSE.TXT)
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os

from GeoEco.DynamicDocString import DynamicDocString
from GeoEco.Internationalization import _
from GeoEco.Logging import Logger

class SpeciesDiversity(object):
    __doc__ = DynamicDocString()

    @classmethod
    def CalculateDiversityIndexForArcGISPolygons(cls, zonesPolygons, index, destField, speciesPoints, speciesIDField, speciesCountField=None, where=None, hurlbertSampleSize=None):
        cls.__doc__.Obj.ValidateMethodInvocation()

        # Perform additional validation.

        if index in [u'hurlbert\'s expected species richness', u'hurlbert\'s standard errors'] and hurlbertSampleSize is None:
            Logger.RaiseException(ValueError(_(u'When calculating %(name)s, you must also specify a sample size.') % {u'name': index[0].upper() + index[1:]}))
        
        try:
            # Create a temporary geodatabase to hold temporary results.

            from GeoEco.DataManagement.Directories import TemporaryDirectory
            tempDir = TemporaryDirectory()

            from GeoEco.ArcGIS import GeoprocessorManager
            gp = GeoprocessorManager.GetWrappedGeoprocessor()

            gp.CreatePersonalGDB_management(tempDir.Path, u'Temp.mdb')
            tempGeoDB = os.path.join(tempDir.Path, u'Temp.mdb')

            try:
                # Create a feature layer that selects points that have a
                # species ID, a species count (if the caller provided a
                # species count field), and that match the caller's where
                # clause (if provided).

                Logger.Info(_(u'Intersecting the points with the polygons.'))

                if where is None:
                    where = speciesIDField + u' IS NOT NULL'
                else:
                    where +=  u' AND ' + speciesIDField + u' IS NOT NULL'

                if speciesCountField is not None:
                    where +=  u' AND ' + speciesCountField + u' IS NOT NULL'
                    
                pointsLayer = GeoprocessorManager.GetUniqueLayerName()
                gp.MakeFeatureLayer_management(speciesPoints, pointsLayer, where)

                # Intersect the points with the polygons.

                try:
                    intersectedPoints = os.path.join(tempGeoDB, u'IntersectedPoints')
                    gp.Intersect_analysis("'" + pointsLayer + u"';'" + zonesPolygons + "'", intersectedPoints)
                    
                finally:
                    gp.Delete_management(pointsLayer)

                # If none of the points intersected the polygons,
                # write zero for the diversity index and return now.

                from GeoEco.DatabaseAccess.ArcGIS import ArcGIS91DatabaseConnection
                conn1 = ArcGIS91DatabaseConnection()

                indexValues = {}
                intersectedPointCount = conn1.GetRowCount(intersectedPoints)
                if intersectedPointCount <= 0:
                    Logger.Warning(_(u'None of the points intersected the polygons. The species diversity index values will be set to zero for all polygons.'))

                else:
                    # For each {Zone, Species} combination, count the
                    # number of occurrences.

                    Logger.Info(_(u'Summing species counts per polygon.'))

                    polygonFIDField = u'FID_' + os.path.basename(zonesPolygons).split(os.path.extsep)[0]
                    if speciesCountField is None:
                        gp.Frequency_analysis(intersectedPoints, os.path.join(tempGeoDB, 'Counts'), polygonFIDField + u';' + speciesIDField)
                    else:
                        gp.Frequency_analysis(intersectedPoints, os.path.join(tempGeoDB, 'Counts'), polygonFIDField + u';' + speciesIDField, speciesCountField)

                    # Read the statistics table into memory and build a
                    # dictionary suitable for passing to R. R will
                    # translate it into a data frame.

                    from GeoEco.DatabaseAccess.InMemory import InMemoryDatabaseConnection
                    conn2 = InMemoryDatabaseConnection()

                    Logger.Info(_(u'Reading species counts into memory.'))

                    speciesCounts = {}
                    if speciesCountField is None:
                        conn2.ImportTable(conn1, os.path.join(tempGeoDB, 'Counts'), u'Counts', [polygonFIDField, speciesIDField, u'FREQUENCY'])
                        speciesCounts['Count'] = conn2.GetFieldValues(u'Counts', u'FREQUENCY')
                    else:
                        conn2.ImportTable(conn1, os.path.join(tempGeoDB, 'Counts'), u'Counts', [polygonFIDField, speciesIDField, speciesCountField])
                        speciesCounts['Count'] = conn2.GetFieldValues(u'Counts', speciesCountField)
                    speciesCounts['PolygonFID'] = conn2.GetFieldValues(u'Counts', polygonFIDField)
                    speciesCounts['SpeciesID'] = conn2.GetFieldValues(u'Counts', speciesIDField)

                    # Calculate diversity using the R vegan package.

                    if index == u'shannon index':
                        veganIndexName = 'shannon'
                    elif index == u'simpson\'s index of diversity':
                        veganIndexName = 'simpson'
                    elif index == u'simpson\'s reciprocal index':
                        veganIndexName = 'invsimpson'
                    elif index not in [u'species richness', u'hurlbert\'s expected species richness', u'hurlbert\'s standard errors']:
                        Logger.RaiseException(RuntimeError(_(u'Programming error in this tool: unknown species diversity index "%s". Please contact the author of this tool for assistance.') % index))

                    from GeoEco.R import R

                    r = R.GetInterpreter()
                    r('library(vegan)')
                    dataFrameName = R.GetUniqueVariableName()
                    r[dataFrameName] = speciesCounts
                    try:
                        if index == u'species richness':
                            result = r('specnumber(xtabs(Count ~ PolygonFID + SpeciesID, %s))' % dataFrameName)
                        elif index == u'hurlbert\'s expected species richness':
                            result = r('rarefy(xtabs(Count ~ PolygonFID + SpeciesID, %s), %i)' % (dataFrameName, hurlbertSampleSize))
                        elif index == u'hurlbert\'s standard errors':
                            result = r('rarefy(xtabs(Count ~ PolygonFID + SpeciesID, %s), %i, se=TRUE)[2,]' % (dataFrameName, hurlbertSampleSize))
                        else:
                            result = r('diversity(xtabs(Count ~ PolygonFID + SpeciesID, %s), "%s")' % (dataFrameName, veganIndexName))
                    finally:
                        r('if (exists("%s")) rm("%s")' % (dataFrameName, dataFrameName))

                    # If there was only one polygon, R / rpy returns a
                    # single float, not a dictionary. Handle this.

                    if isinstance(result, dict):
                        indexValues = result
                    elif len(speciesCounts['PolygonFID']) > 0 and len(set(speciesCounts['PolygonFID'])) == 1:
                        indexValues[str(speciesCounts['PolygonFID'][0])] = result
                    else:
                        Logger.RaiseException(RuntimeError(_(u'Programming error in this tool: unknown value returned by the R vegan package. Please contact the author of this tool for assistance. The Python representation of the returned values was: %r') % result))

                # Write the index values to the polygons.

                Logger.Info(_(u'Writing species diversity index values to the polygons.'))
                
                fidField = gp.Describe(zonesPolygons).OIDFieldName
                cur = conn1.OpenUpdateCursor(zonesPolygons)
                try:
                    while cur.NextRow():
                        fidStr = str(cur.GetValue(fidField))
                        if indexValues.has_key(fidStr):
                            cur.SetValue(destField, indexValues[fidStr])
                        else:
                            cur.SetValue(destField, 0)
                        cur.UpdateRow()
                finally:
                    del cur

            finally:
                # Delete the temporary personal geodatabase now. If we do
                # not explicitly delete it here, the geoprocessor will
                # keep a handle open through MS Access APIs, and the
                # temporary directory will not be able to be deleted.

                try:
                    gp.Delete_management(tempGeoDB)
                except:
                    pass

        except:
            Logger.LogExceptionAsError()
            raise
        
        return zonesPolygons

###############################################################################
# Metadata: module
###############################################################################

from GeoEco.ArcGIS import ArcGISDependency, ArcGISProductDependency
from GeoEco.Metadata import *
from GeoEco.R import RDependency, RPackageDependency
from GeoEco.Types import *

AddModuleMetadata(shortDescription=_(u'Provides methods for analyzing species diversity.'))

###############################################################################
# Metadata: SpeciesDiversity class
###############################################################################

AddClassMetadata(SpeciesDiversity,
    shortDescription=_(u'Provides methods for analyzing species diversity.'),
    isExposedAsCOMServer=True,
    comIID=u'{D2F1B026-AC57-4CDA-8962-2C3945BF2625}',
    comCLSID=u'{20B7C22C-D593-440A-8E49-EE1F745BC6C6}')

# Public method: SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons

AddMethodMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons,
    shortDescription=_(u'Given polygons representing zones of interest and points representing species occurrence observations, calculates a species diversity index for each polygon.'),
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Calculate Species Diversity Index for Polygons'),
    arcGISToolCategory=_(u'Spatial and Temporal Analysis\\Analyze Species Diversity'),
    dependencies=[ArcGISDependency(9, 1), ArcGISProductDependency(u'ArcInfo'), RDependency(2, 5, 0), RPackageDependency(u'vegan')])

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'cls',
    typeMetadata=ClassOrClassInstanceTypeMetadata(cls=SpeciesDiversity),
    description=_(u'%s class or an instance of it.') % SpeciesDiversity.__name__)

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'zonesPolygons',
    typeMetadata=ArcGISFeatureLayerTypeMetadata(mustExist=True, allowedShapeTypes=[u'Polygon']),
    description=_(
u"""Polygon features representing the zones of interest for which the
species diversity index should be calculated."""),
    arcGISDisplayName=_(u'Polygons representing zones of interest'))

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'index',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Hurlbert\'s expected species richness', u'Hurlbert\'s standard errors', u'Shannon index', u'Simpson\'s index of diversity', u'Simpson\'s reciprocal index', u'Species richness'], makeLowercase=True),
    description=_(
u"""Species diversity index to calculate, one of:

* Hurlbert's expected species richness - sometimes abbreviated ES(n)
  in the literature. Gives the expected species richness in random
  subsamples from the community. To use this index, you must also
  specify the sample size parameter (n). This parameter should be
  smaller than total community size, but the function will silently
  work for larger sizes as well. The calculation is based on Hurlbert
  (1971).

* Hurlbert's standard errors - Gives the standard errors for
  Hurlbert's expected species richness (see above). You must also
  specify the sample size parameter. The standard errors are based on
  Heck et al. (1975).

* Shannon index - abbreviated H in the literature. Also known as the
  Shannon-Wiener or Shannon-Weaver index. The calculation is performed
  using a natural logarithm.

* Simpson's index of diversity - abbreviated 1-D in the literature.

* Simpson's reciprocal index - abbreviated 1/D in the literature.

* Species richness - abbreviated S in the literature. The simplest
  measure of diversity: the count of distinct species present in the
  zone.

Please consult the literature for exact definitions of the indices.
The calculations are performed by the R vegan package."""),
    arcGISDisplayName=_(u'Species diversity index'))

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'destField',
    typeMetadata=ArcGISFieldTypeMetadata(mustExist=True, allowedFieldTypes=[u'FLOAT', u'DOUBLE']),
    description=_(
u"""Field of the polygon features to receive the value of the species
diversity index. The field must already exist."""),
    arcGISParameterDependencies=[u'zonesPolygons'],
    arcGISDisplayName=_(u'Destination field'))

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'speciesPoints',
    typeMetadata=ArcGISFeatureLayerTypeMetadata(mustExist=True, allowedShapeTypes=[u'Point']),
    description=_(
u"""Point features representing species occurrence observations."""),
    arcGISDisplayName=_(u'Points representing species occurrence observations'))

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'speciesIDField',
    typeMetadata=ArcGISFieldTypeMetadata(mustExist=True, allowedFieldTypes=[u'TEXT', u'SHORT', u'LONG']),
    description=_(
u"""Field of the point features that specifies the name or ID of the
species that was observed.

It does not matter how species names or IDs are coded in this field.
The tool simply counts occurrences of the different values and does
not care what they actually mean."""),
    arcGISParameterDependencies=[u'speciesPoints'],
    arcGISDisplayName=_(u'Species ID field'))

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'speciesCountField',
    typeMetadata=ArcGISFieldTypeMetadata(canBeNone=True, mustExist=True, allowedFieldTypes=[u'SHORT', u'LONG']),
    description=_(
u"""Field of the point features that specifies the number of
individuals observed.

This field is optional. If not specified, the tool assumes that one
individual was observed at each point."""),
    arcGISParameterDependencies=[u'speciesPoints'],
    arcGISDisplayName=_(u'Species count field'))

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'where',
    typeMetadata=SQLWhereClauseTypeMetadata(canBeNone=True),
    description=_(
u"""SQL WHERE clause expression that specifies the subset of points to
use when calculating the species diversity index. If this parameter is
not provided, all of the points will be used.

The exact syntax of this expression depends on the type of feature
class that contains the point features. ESRI recommends you reference
fields using the following syntax:

* If the points are stored in an ArcInfo coverage or a shapefile,
  enclose field names in double quotes in the SQL expression:
  "MY_FIELD".

* If the points are stored in a personal geodatabase (.mdb file),
  enclose field names in square brackets: [MY_FIELD].

* If the point features are stored in an ArcSDE or ArcIMS feature
  class, don't enclose field names: MY_FIELD."""),
    arcGISDisplayName=_(u'Where clause'),
    arcGISParameterDependencies=[u'speciesPoints'])

AddArgumentMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'hurlbertSampleSize',
    typeMetadata=IntegerTypeMetadata(canBeNone=True, minValue=2),
    description=_(
u"""Sample size parameter for the calculation of Hurlbert's expected
species richness.

You must provide this parameter when claculating Hurlbert's expected
species richness. This parameter is ignored when calculating other
indices."""),
    arcGISDisplayName=_(u'Hurlbert\'s sample size'),
    arcGISParameterDependencies=[u'speciesPoints'])

AddResultMetadata(SpeciesDiversity.CalculateDiversityIndexForArcGISPolygons, u'updatedZonesPolygons',
    typeMetadata=ArcGISFeatureLayerTypeMetadata(),
    description=_(u'Updated zones polygons.'),
    arcGISDisplayName=_(u'Updated zones polygons'),
    arcGISParameterDependencies=[u'zonesPolygons'])

###############################################################################
# Names exported by this module
###############################################################################

__all__ = ['SpeciesDiversity']
